@@ -144,6 +144,9 @@ void FileProtocol::copy(const QUrl &srcUrl, const QUrl &destUrl,
         return;
     }
 
+    // nobody shall be allowed to peek into the file during creation
+    dest_file.setPermissions(QFileDevice::ReadOwner | QFileDevice::WriteOwner);
+
 #if HAVE_FADVISE
     posix_fadvise(dest_file.handle(), 0, 0, POSIX_FADV_SEQUENTIAL);
 #endif
@@ -153,7 +156,7 @@ void FileProtocol::copy(const QUrl &srcUrl, const QUrl &destUrl,
     if (acl && !isExtendedACL(acl)) {
         // qDebug() << _dest.data() << " doesn't have extended ACL";
         acl_free(acl);
-        acl = NULL;
+        acl = nullptr;
     }
 #endif
     totalSize(buff_src.st_size);
@@ -188,7 +191,6 @@ void FileProtocol::copy(const QUrl &srcUrl, const QUrl &destUrl,
                 // qDebug() << "sendfile() error:" << strerror(errno);
                 if (errno == ENOSPC) { // disk full
                     error(KIO::ERR_DISK_FULL, dest);
-                    remove(_dest.data());
                 } else {
                     error(KIO::ERR_SLAVE_DEFINED,
                           i18n("Cannot copy file from %1 to %2. (Errno: %3)",
@@ -204,6 +206,7 @@ void FileProtocol::copy(const QUrl &srcUrl, const QUrl &destUrl,
                 acl_free(acl);
             }
 #endif
+            dest_file.remove();  // don't keep partly copied file
             return;
         }
         if (n == 0) {
@@ -215,9 +218,8 @@ void FileProtocol::copy(const QUrl &srcUrl, const QUrl &destUrl,
             if (dest_file.write(buffer, n) != n) {
                 if (dest_file.error() == QFileDevice::ResourceError) {  // disk full
                     error(KIO::ERR_DISK_FULL, dest);
-                    remove(_dest.data());
                 } else {
-                    qWarning() << "Couldn't write[2]. Error:" << dest_file.errorString();
+                    qCWarning(KIO_FILE) << "Couldn't write[2]. Error:" << dest_file.errorString();
                     error(KIO::ERR_CANNOT_WRITE, dest);
                 }
 #if HAVE_POSIX_ACL
@@ -225,6 +227,7 @@ void FileProtocol::copy(const QUrl &srcUrl, const QUrl &destUrl,
                     acl_free(acl);
                 }
 #endif
+                dest_file.remove();  // don't keep partly copied file
                 return;
             }
             processed_size += n;
@@ -238,28 +241,32 @@ void FileProtocol::copy(const QUrl &srcUrl, const QUrl &destUrl,
     dest_file.close();
 
     if (dest_file.error() != QFile::NoError) {
-        qWarning() << "Error when closing file descriptor[2]:" << dest_file.errorString();
+        qCWarning(KIO_FILE) << "Error when closing file descriptor[2]:" << dest_file.errorString();
         error(KIO::ERR_CANNOT_WRITE, dest);
 #if HAVE_POSIX_ACL
         if (acl) {
             acl_free(acl);
         }
 #endif
+        dest_file.remove();  // don't keep partly copied file
         return;
     }
 
     // set final permissions
-    if (_mode != -1) {
-        if ((::chmod(_dest.data(), _mode) != 0)
+    // if no special mode given, preserve the mode from the sourcefile
+    if (_mode == -1) {
+        _mode = buff_src.st_mode;
+    }
+
+    if ((::chmod(_dest.data(), _mode) != 0)
 #if HAVE_POSIX_ACL
-                || (acl && acl_set_file(_dest.data(), ACL_TYPE_ACCESS, acl) != 0)
+        || (acl && acl_set_file(_dest.data(), ACL_TYPE_ACCESS, acl) != 0)
 #endif
-           ) {
-            KMountPoint::Ptr mp = KMountPoint::currentMountPoints().findByPath(dest);
-            // Eat the error if the filesystem apparently doesn't support chmod.
-            if (mp && mp->testFileSystemFlag(KMountPoint::SupportsChmod)) {
-                warning(i18n("Could not change permissions for\n%1", dest));
-            }
+       ) {
+        KMountPoint::Ptr mp = KMountPoint::currentMountPoints().findByPath(dest);
+        // Eat the error if the filesystem apparently doesn't support chmod.
+        if (mp && mp->testFileSystemFlag(KMountPoint::SupportsChmod)) {
+            warning(i18n("Could not change permissions for '%1'", dest));
         }
     }
 #if HAVE_POSIX_ACL
@@ -268,12 +275,21 @@ void FileProtocol::copy(const QUrl &srcUrl, const QUrl &destUrl,
     }
 #endif
 
+    // preserve ownership
+    if (::chown(_dest.data(), -1 /*keep user*/, buff_src.st_gid) == 0) {
+        // as we are the owner of the new file, we can always change the group, but
+        // we might not be allowed to change the owner
+        (void)::chown(_dest.data(), buff_src.st_uid, -1 /*keep group*/);
+    } else {
+        qCWarning(KIO_FILE) << QStringLiteral("Couldn't preserve group for '%1'").arg(dest);
+    }
+
     // copy access and modification time
     struct utimbuf ut;
     ut.actime = buff_src.st_atime;
     ut.modtime = buff_src.st_mtime;
     if (::utime(_dest.data(), &ut) != 0) {
-        qWarning() << QString::fromLatin1("Couldn't preserve access and modification time for\n%1").arg(dest);
+        qCWarning(KIO_FILE) << QStringLiteral("Couldn't preserve access and modification time for '%1'").arg(dest);
     }
 
     processedSize(buff_src.st_size);
@@ -312,7 +328,7 @@ void FileProtocol::listDir(const QUrl &url)
     const QString path(url.toLocalFile());
     const QByteArray _path(QFile::encodeName(path));
     DIR *dp = opendir(_path.data());
-    if (dp == 0) {
+    if (dp == nullptr) {
         switch (errno) {
         case ENOENT:
             error(KIO::ERR_DOES_NOT_EXIST, path);
@@ -356,7 +372,7 @@ void FileProtocol::listDir(const QUrl &url)
     QT_STATBUF st;
 #endif
     QT_DIRENT *ep;
-    while ((ep = QT_READDIR(dp)) != 0) {
+    while ((ep = QT_READDIR(dp)) != nullptr) {
         entry.clear();
 
         const QString filename = QFile::decodeName(ep->d_name);
@@ -389,7 +405,7 @@ void FileProtocol::listDir(const QUrl &url)
             if (isSymLink) {
                 // for symlinks obey the UDSEntry contract and provide UDS_LINK_DEST
                 // even if we don't know the link dest (and DeleteJob doesn't care...)
-                entry.insert(KIO::UDSEntry::UDS_LINK_DEST, QLatin1String("Dummy Link Target"));
+                entry.insert(KIO::UDSEntry::UDS_LINK_DEST, QStringLiteral("Dummy Link Target"));
             }
             listEntry(entry);
 
@@ -451,7 +467,7 @@ void FileProtocol::rename(const QUrl &srcUrl, const QUrl &destUrl,
         if ((errno == EACCES) || (errno == EPERM)) {
             error(KIO::ERR_ACCESS_DENIED, dest);
         } else if (errno == EXDEV) {
-            error(KIO::ERR_UNSUPPORTED_ACTION, QLatin1String("rename"));
+            error(KIO::ERR_UNSUPPORTED_ACTION, QStringLiteral("rename"));
         } else if (errno == EROFS) { // The file is on a read-only filesystem
             error(KIO::ERR_CANNOT_DELETE, src);
         } else {
@@ -478,6 +494,7 @@ void FileProtocol::symlink(const QString &target, const QUrl &destUrl, KIO::JobF
                 }
                 // Try again - this won't loop forever since unlink succeeded
                 symlink(target, destUrl, flags);
+                return;
             } else {
                 QT_STATBUF buff_dest;
                 if (QT_LSTAT(QFile::encodeName(dest).constData(), &buff_dest) == 0 && ((buff_dest.st_mode & QT_STAT_MASK) == QT_STAT_DIR)) {
@@ -524,7 +541,7 @@ void FileProtocol::del(const QUrl &url, bool isfile)
          *****/
 
         // qDebug() << "Deleting directory " << url;
-        if (metaData(QLatin1String("recurse")) == QLatin1String("true")) {
+        if (metaData(QStringLiteral("recurse")) == QLatin1String("true")) {
             if (!deleteRecursive(path)) {
                 return;
             }
@@ -595,7 +612,7 @@ void FileProtocol::chown(const QUrl &url, const QString &owner, const QString &g
 
 void FileProtocol::stat(const QUrl &url)
 {
-    if (!url.isLocalFile()) {
+    if (!isLocalFileSameHost(url)) {
         QUrl redir(url);
         redir.setScheme(config()->readEntry("DefaultRemoteProtocol", "smb"));
         redirection(redir);
